#pragma once
#include <inttypes.h>
#include <vector>
#include <unordered_set>
#include <stdexcept>
#include <utility>
#include <map>
#include <string>

namespace lidar_tracker
{
  // enum ObservationType : int
  // {
  //   OT_UNKNOWN = 0,
  //   OT_VEHICLE = 1,
  //   OT_TRUCK = 2,
  //   OT_CYCLIST = 3,
  //   OT_PEDESTRIAN = 4,
  // };

  enum ObservationType : int
  {
    car = 0,
    truck = 1,
    trailer = 2,
    bus = 3,
    construction_vehicle = 4,
    bicycle = 5,
    motorcycle = 6,
    pedestrain = 7,
    traffic_cone = 8,
    barrier = 9,
  };

  const std::map <int, std::string> ObservationTypeToString = {
    {car, "car"},
    {truck, "truck"},
    {trailer, "trailer"},
    {bus, "bus"},
    {construction_vehicle, "construction_vehicle"},
    {bicycle, "bicycle"},
    {motorcycle, "motorcycle"},
    {pedestrain, "pedestrain"},
    {traffic_cone, "traffic_cone"},
    {barrier, "barrier"},
  };

  enum TrackLifeState : int
  {
    TLS_IDLE = 0,
    TLS_INIT = 1,
    TLS_CONFIRMED = 2,
    TLS_LOST = 3,
  };

  template <typename T>
  class Optional
  {
  public:
    Optional() : has_value_(false) {}

    Optional(const T &value) : has_value_(true), value_(value) {}

    Optional(T &&value) : has_value_(true), value_(std::move(value)) {}

    bool has_value() const { return has_value_; }
    T value() const { return value_;}

  private:
    bool has_value_;
    T value_;
  };

  class LidarObservationProto
  {
  public:
    LidarObservationProto() = default;
    double center_x() const { return center_x_; }
    double center_y() const { return center_y_; }
    double center_z() const { return center_z_; }
    double timestamp() const { return timestamp_; }
    double length() const { return length_; }
    double width() const { return width_; }
    double height() const { return height_; }
    double yaw() const { return yaw_; }
    ObservationType type() const { return type_; }

    void set_timestamp(double timestamp) { timestamp_ = timestamp; }
    void set_center_x(double center_x) { center_x_ = center_x; }
    void set_center_y(double center_y) { center_y_ = center_y; }
    void set_center_z(double center_z) { center_z_ = center_z; }
    void set_length(double length) { length_ = length; }
    void set_width(double width) { width_ = width; }
    void set_height(double height) { height_ = height; }
    void set_yaw(double yaw) { yaw_ = yaw; }
    void set_type(ObservationType type) { type_ = type; }

  private:
    double timestamp_;
    double center_x_;
    double center_y_;
    double center_z_;
    double length_;
    double width_;
    double height_;
    double yaw_;
    ObservationType type_;
  };

  class ObservationProto
  {
  public:
    ObservationProto() = default;

    double timestamp() const { return timestamp_; }
    void set_timestamp(double timestamp) { timestamp_ = timestamp; }
    void set_id(int32_t id) { id_ = id; }
    void set_type(ObservationType type) { type_ = type; }

    const LidarObservationProto &lidar_observation() const
    {
      return lidar_observation_;
    }
    LidarObservationProto *mutable_lidar_observation() { return &lidar_observation_; }

  private:
    double timestamp_;
    int32_t id_;
    ObservationType type_;
    LidarObservationProto lidar_observation_;
  };

  class ObservationsProto
  {
  public:
    ObservationsProto() = default;
    int observations_size() const { return observations_.size(); }
    const std::vector<ObservationProto> &observations() const { return observations_; }
    double min_timestamp() const { return min_timestamp_.value(); };
    bool has_min_timestamp() const { return min_timestamp_.has_value();}
    void set_min_timestamp(double min_timestamp) { min_timestamp_ = min_timestamp; }
    ObservationProto *add_observations()
    {
      ObservationProto obs;
      observations_.push_back(obs);
      return &observations_[observations_.size() - 1];
    }

  private:
    std::vector<ObservationProto> observations_;
    Optional<double> min_timestamp_;
  };
}